PRINT 'CREATING DEPENDDB DELETE (Delete Procedures and Functions)'

-- Drop procedures
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'DeleteDependenciesByDependentAssembly') AND type in (N'P')) DROP PROCEDURE DeleteDependenciesByDependentAssembly;
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'DeleteProjectAssemblyAssociations') AND type in (N'P')) DROP PROCEDURE DeleteProjectAssemblyAssociations;
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'DeleteProjectVersion') AND type in (N'P')) DROP PROCEDURE DeleteProjectVersion;
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'DeleteProjectVersions') AND type in (N'P')) DROP PROCEDURE DeleteProjectVersions;
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'DeleteProjectBranch') AND type in (N'P')) DROP PROCEDURE DeleteProjectBranch;
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'DeleteProject') AND type in (N'P')) DROP PROCEDURE DeleteProject;
GO

-- Drop functions
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'CanRequestProjectVersionDeletion') AND type in (N'FN', N'IF', N'TF', N'FS', N'FT')) DROP FUNCTION CanRequestProjectVersionDeletion;
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'CanDeleteProjectBranch') AND type in (N'FN', N'IF', N'TF', N'FS', N'FT')) DROP FUNCTION CanDeleteProjectBranch;
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'CanDeleteProject') AND type in (N'FN', N'IF', N'TF', N'FS', N'FT')) DROP FUNCTION CanDeleteProject;
GO

-- Drop types
IF TYPE_ID(N'ListOfInt') IS NOT NULL DROP TYPE ListOfInt
GO

-- ListOfInt
CREATE TYPE ListOfInt AS TABLE (n INT NOT NULL)
GO

-- DeleteDependenciesByDependentAssembly
-- Procedure: Delete dependencies by dependent assembly
CREATE PROCEDURE DeleteDependenciesByDependentAssembly
	@assemblyId INT
AS
BEGIN
	RAISERROR ('Deleting Dependencies with DepedentParts that belong to AssemblyId %d', 10, 1, @assemblyId) WITH NOWAIT

	DELETE FROM Dependency
	WHERE DependentPartId IN (
		SELECT Id
		FROM DependencyPart
		WHERE AssemblyId = @assemblyId
	)
END
GO

-- DeleteProjectAssemblyAssociations
-- Procedure: Deletes every entry in ProjectVersionAssembly that has the given @projectVersionId
CREATE PROCEDURE DeleteProjectAssemblyAssociations
	@projectVersionId INT
AS
BEGIN
	RAISERROR ('Deleting ProjectVersionAssemblies with ProjectVersionId %d', 10, 1, @projectVersionId) WITH NOWAIT

	DELETE FROM ProjectVersionAssembly WHERE ProjectVersionId = @projectVersionId
END
GO

-- DeleteProjectVersions
-- Procedure: Deletes the given project versions, all their ProjectVersionAssembly entries
-- and all assemblies which are exclusively used in those project versions.
-- (including their parts/dependencies/types/members)
CREATE PROCEDURE DeleteProjectVersions 
	@projectVersionIds ListOfInt READONLY
AS
BEGIN
	-- Step 1: Delete all ProjectVersionAssembly entries
	DELETE FROM ProjectVersionAssembly WHERE ProjectVersionId IN (
		SELECT n FROM @projectVersionIds
	)
	
	-- Step 2: Delete project versions
	DELETE FROM ProjectVersion WHERE Id IN (
		SELECT n FROM @projectVersionIds
	)
	
	-- Step 3: Delete all dependencies where the dependent part's assembly
	-- is not part of any project version (i.e. is going to be deleted)
	DELETE FROM Dependency WHERE DependentPartId IN (
		SELECT dp.Id FROM DependencyPart dp
		INNER JOIN Assembly da ON da.Id = dp.AssemblyId
		WHERE NOT EXISTS (
			SELECT * FROM ProjectVersionAssembly pva
			WHERE pva.AssemblyId = da.Id
		)
	)
	
	-- Step 4: Delete all dependency parts that are not part of a
	-- dependency anymore
	DELETE p FROM DependencyPart p WHERE NOT EXISTS (
		SELECT * FROM Dependency d
		WHERE d.DependentPartId = p.Id OR d.ReferencedPartId = p.Id
	)
	
	-- Step 5: Delete all types/members which are not in use anymore
	DELETE t FROM [Type] t WHERE NOT EXISTS (
		SELECT * FROM DependencyPart dp
		WHERE dp.TypeId = t.Id
	)
	
	DELETE m FROM Member m WHERE NOT EXISTS (
		SELECT * FROM DependencyPart dp
		WHERE dp.MemberId = m.Id
	)
	
	-- Step 6: Delete all assemblies which are not assigned to any
	-- project version anymore
	DELETE a FROM Assembly a WHERE NOT EXISTS (
		SELECT * FROM ProjectVersionAssembly pva
		WHERE pva.AssemblyId = a.Id
	)
END
GO

-- DeleteProjectVersion
-- Procedure: See DeleteProjectVersions.
CREATE PROCEDURE DeleteProjectVersion
	@projectVersionId INT
AS
BEGIN
	DECLARE @projectVersionIds ListOfInt
	INSERT @projectVersionIds (n) VALUES (@projectVersionId)
	EXEC dbo.DeleteProjectVersions @projectVersionIds
END
GO

-- DeleteProjectBranch
-- Procedure: Deletes a project branch and all its project versions
CREATE PROCEDURE DeleteProjectBranch
	@branchId INT
AS
BEGIN
	-- Step 1: Delete all associated project versions
	--DECLARE @versionIds ListOfInt
	--INSERT @versionIds
	--	SELECT Id
	--	FROM ProjectVersion 
	--	WHERE ProjectBranchId = @branchId
		
	--EXEC dbo.DeleteProjectVersions @versionIds
	
	-- Step 2: Delete the project branch
	DELETE FROM ProjectBranch WHERE Id = @branchId
END
GO


-- DeleteProject
-- Procedure: Deletes a project and all its project branches
CREATE PROCEDURE DeleteProject
	@projectId INT
AS
BEGIN
	-- Step 1: Delete all associated project branches
	DECLARE @branchIds CURSOR
	DECLARE @branchId INT
	
	SET @branchIds =
		CURSOR FOR
		SELECT Id
		FROM ProjectBranch
		WHERE ProjectId = @projectId
		
	OPEN @branchIds
	FETCH NEXT FROM @branchIds INTO @branchId
	
	WHILE @@FETCH_STATUS = 0 BEGIN
		EXEC dbo.DeleteProjectBranch @branchId
		
		FETCH NEXT FROM @branchIds INTO @branchId
	END
	
	CLOSE @branchIds
	DEALLOCATE @branchIds
	
	-- Step 2: Delete the project
	DELETE FROM Project WHERE Id = @projectId
END
GO

-- CanDeleteProjectVersion
-- Function: Returns whether a project version can be deleted.
CREATE FUNCTION CanRequestProjectVersionDeletion ( @projectVersionId INT )
RETURNS BIT
AS
BEGIN
	-- Check if there is any dependency on the project version
	-- in an assembly which is exclusively associated with the project version
	IF (SELECT CAST(IsPinned AS INT) + CAST(DeleteRequested AS INT) FROM ProjectVersion WHERE Id = @projectVersionId) > 0
		RETURN 0
	
	RETURN 1
END
GO

-- CanDeleteProjectBranch
-- Function: Returns whether a project branch can be deleted.
CREATE FUNCTION CanDeleteProjectBranch ( @projectBranchId INT )
RETURNS BIT
AS
BEGIN
	-- Check if there are still versions in the branch
	IF EXISTS (
		SELECT * FROM ProjectVersion
		WHERE ProjectBranchId = @projectBranchId
	) RETURN 0
	
	RETURN 1
END
GO

-- CanDeleteProject
-- Function: Returns whether a project can be deleted.
CREATE FUNCTION CanDeleteProject ( @projectId INT )
RETURNS BIT
AS
BEGIN
	-- Check if there are still branches in the project
	IF EXISTS (
		SELECT * FROM ProjectBranch
		WHERE ProjectId = @projectId
		AND dbo.CanDeleteProjectBranch(Id) = 0
	) RETURN 0
	
	RETURN 1
END
GO
