PRINT 'CREATING DEPENDDB DATABASE RELEVANCY (Database Procedures and Functions)'
GO

-- Drop procedures
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'PropagateRelevancyStatus') AND type in (N'P')) DROP PROCEDURE PropagateRelevancyStatus;
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'RebuildRelevancyStatus') AND type in (N'P')) DROP PROCEDURE RebuildRelevancyStatus;
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'RebuildQueryRelevancyLevel') AND type in (N'P')) DROP PROCEDURE RebuildQueryRelevancyLevel;
GO

-- Drop functions
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'IsProjectVersionLatestForAssembly') AND type in (N'FN', N'IF', N'TF', N'FS', N'FT')) DROP FUNCTION IsProjectVersionLatestForAssembly;
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'RelevancyStatus_RelevantForQueries') AND type in (N'FN', N'IF', N'TF', N'FS', N'FT')) DROP FUNCTION RelevancyStatus_RelevantForQueries;
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'RelevancyStatus_NonCleanupable') AND type in (N'FN', N'IF', N'TF', N'FS', N'FT')) DROP FUNCTION RelevancyStatus_NonCleanupable;
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'RelevancyStatus_Cleanupable') AND type in (N'FN', N'IF', N'TF', N'FS', N'FT')) DROP FUNCTION RelevancyStatus_Cleanupable;
GO

-- Relevant for queries, of course not ready for cleanup
CREATE FUNCTION RelevancyStatus_RelevantForQueries()
RETURNS TINYINT
AS
BEGIN
	RETURN 1
END
GO

-- Not relevant for queries, however, not ready for cleanup (e.g. pinned, retention time not expired, etc.)
CREATE FUNCTION RelevancyStatus_NonCleanupable()
RETURNS TINYINT
AS
BEGIN
	RETURN 2
END
GO

-- Not relevant for queries and ready for cleanup
CREATE FUNCTION RelevancyStatus_Cleanupable()
RETURNS TINYINT
AS
BEGIN
	RETURN 3
END
GO


-- IsProjectVersionLatestForAssembly
-- Function: Determines whether a given PV (@projectVersionId) is the latest PV holding
-- this assembly (@assemblyId).
CREATE FUNCTION IsProjectVersionLatestForAssembly ( @projectVersionId BIGINT, @assemblyId BIGINT )
RETURNS BIT AS
BEGIN
	DECLARE @latestProjectVersionForAssembly BIGINT
	SELECT TOP 1 @latestProjectVersionForAssembly = pv.ID
	FROM Assembly a
	INNER JOIN ProjectVersionAssembly pva ON pva.AssemblyId = a.ID
	INNER JOIN ProjectVersion pv ON pv.ID = pva.ProjectVersionID
	WHERE a.ID = @assemblyId
	-- Alle PV aus demselben Branch
	AND pv.ProjectBranchID = (SELECT ipv.PRojectBranchID FROM ProjectVersion ipv WHERE ipv.ID = @projectVersionId)
	ORDER BY pv.Version DESC
	
	IF @latestProjectVersionForAssembly = @projectVersionId BEGIN
		RETURN 1
	END
	
	RETURN 0
END
GO

-- PropagateRelevancyStatus
-- Procedure: Propagates the ProjectVersion.RelevancyStatus flag to all directly or
-- transitively referenced PVs if they do not have another RelevancyStatus already.
CREATE PROCEDURE PropagateRelevancyStatus
	@relevancyStatus TINYINT
AS
BEGIN
	DECLARE @iteration INT = 1

	DECLARE @running BIT = 1
	WHILE @running = 1 BEGIN
		RAISERROR ('PropagateRelevancyStatus(%d) starting iteration %d', 10, 1, @relevancyStatus, @iteration) WITH NOWAIT
		
		-- Set all project versions that are referenced by project versions with the given relevancy status to the same relevancy status
		; WITH ReferencedByVersionsWithRelevancyStatus AS (
			SELECT rpv.Id as ReferencedProjectVersionId, MAX(dpv.ID) as DependentProjectVersionId
			FROM ProjectVersion dpv
			INNER JOIN ProjectVersionAssembly dpva ON dpva.ProjectVersionId = dpv.Id		
			INNER JOIN Assembly da ON da.Id = dpva.AssemblyId
			INNER JOIN DependencyPart dp ON dp.AssemblyID = da.ID
			INNER JOIN Dependency d ON d.DependentPartID = dp.ID
			INNER JOIN DependencyPart rp ON rp.ID = d.ReferencedPartID
			INNER JOIN Assembly ra ON ra.ID = rp.AssemblyID
			INNER JOIN ProjectVersionAssembly rpva ON rpva.AssemblyId = ra.Id
			INNER JOIN ProjectVersion rpv ON rpv.Id = rpva.ProjectVersionId
			WHERE dpv.RelevancyStatus = @relevancyStatus
			AND dbo.IsProjectVersionLatestForAssembly(rpv.ID, ra.ID) = 1
			AND rpv.RelevancyStatus IS NULL
			GROUP BY rpv.ID
		) 
		UPDATE pv
		SET RelevancyStatus = @relevancyStatus, RelevancyStatusSourceProjectVersion = r.DependentProjectVersionId
		FROM ProjectVersion pv
		INNER JOIN ReferencedByVersionsWithRelevancyStatus r ON r.ReferencedProjectVersionId = pv.ID
		
		-- Execute statement as long as rows have been affected
		IF @@ROWCOUNT = 0 BEGIN
			SET @running = 0
		END
		
		SET @iteration = @iteration + 1
	END
END
GO

-- RebuildQueryRelevancyLevel
-- Procedure: Completely rebuild the Dependency.QueryRelevancy level
CREATE PROCEDURE RebuildQueryRelevancyLevel AS
BEGIN
	SET NOCOUNT ON

	-- Step 1: Reset the 'QueryRelevancy' for all dependencies to NULL
	UPDATE Dependency SET QueryRelevancy = NULL
	RAISERROR ('Step 1: Reset the QueryRelevancy level for all %d dependencies to NULL.', 10, 1, @@ROWCOUNT) WITH NOWAIT

	-- Step 2: Set QueryRelevancy to Level=1 for all dependencies with the dependent part in a latest PV
	UPDATE d SET QueryRelevancy = 1 FROM Dependency d WHERE EXISTS (
		SELECT *
		FROM DependencyPart idp
		INNER JOIN Assembly a ON a.ID = idp.AssemblyID
		INNER JOIN ProjectVersionAssembly pva ON pva.AssemblyID = a.ID
		INNER JOIN ProjectVersion pv ON pv.ID = pva.ProjectVersionID
		WHERE pv.RelevancyStatus = dbo.RelevancyStatus_RelevantForQueries()
		AND pv.RelevancyStatusSourceProjectVersion IS NULL
		AND idp.ID = d.DependentPartID
	)
	RAISERROR ('Step 2: Set the QueryRelevancy level to 1 for all %d dependencies with the dependent part in a latest PV.', 10, 1, @@ROWCOUNT) WITH NOWAIT

	-- Step 3: Propagate QueryRelevancy
	RAISERROR ('Step 3: Propagating QueryRelevancy...', 10, 1) WITH NOWAIT
	DECLARE @level INT = 1
	DECLARE @running BIT = 1

	WHILE @running = 1 BEGIN

		DECLARE @oldLevel INT = @level
		SET @level = @level + 1
		
		RAISERROR ('  Propagating level: %d...', 10, 1, @level) WITH NOWAIT

		UPDATE d SET QueryRelevancy = @level
		FROM Dependency d
		INNER JOIN DependencyType dt ON dt.ID = d.DependencyTypeID
		INNER JOIN DependencyCategory dc ON dc.ID = dt.Category
		WHERE d.DependentPartID IN (
			SELECT d2.ReferencedPartID FROM Dependency d2
			WHERE d2.QueryRelevancy = @oldLevel
		)
		AND dc.TraverseIndirectDependencies = 1
		AND d.QueryRelevancy IS NULL

		-- Execute statement as long as rows have been affected
		DECLARE @rowCount INT = @@ROWCOUNT
		RAISERROR ('  Propagated level %d to %d dependencies.', 10, 1, @level, @rowCount) WITH NOWAIT
		
		IF @rowCount = 0 BEGIN
			SET @running = 0
		END
	END

	RAISERROR ('Propagating done.', 10, 1) WITH NOWAIT
	
	SET NOCOUNT OFF
END
GO

-- RebuildRelevancyStatus
-- Procedure: Completely rebuilds the ProjectVersion.RelevancyStatus and update the
-- Dependency.QueryRelevancy level accordingly.
CREATE PROCEDURE RebuildRelevancyStatus AS
BEGIN
	RAISERROR ('Rebuilding relevancy status', 10, 1) WITH NOWAIT
	
	-- Step 1: Reset flag 'RelevancyStatus' for all project versions to NULL
	RAISERROR ('Step 1: Reset flag ''RelevancyStatus'' for all project versions to NULL', 10, 1) WITH NOWAIT
	UPDATE ProjectVersion SET RelevancyStatus = NULL, RelevancyStatusSourceProjectVersion = NULL
	
	-- Step 2: Determine latest PVs, set RS on those PVs to RelevantForQueries
	RAISERROR ('Step 2: Determine latest PVs, set RS on those PVs to RelevantForQueries', 10, 1) WITH NOWAIT
	UPDATE ProjectVersion SET RelevancyStatus = dbo.RelevancyStatus_RelevantForQueries() 
	WHERE dbo.ProjectVersionIsLatestInBranch(Id) = 1
		
	-- Step 3: Determine all PVs that are referenced by relevant PVs, set RS
	--         on those PVs to RelevantForQueries as well
	RAISERROR ('Step 3: Determine all PVs that are referenced by relevant PVs, set RS on those PVs to RelevantForQueries as well', 10, 1) WITH NOWAIT
	DECLARE @relevancyStatus TINYINT = dbo.RelevancyStatus_RelevantForQueries()
	EXEC dbo.PropagateRelevancyStatus @relevancyStatus
	
	-- Step 4: Determine which PVs are non-cleanupable
	RAISERROR ('Step 4: Determine which PVs are non-cleanupable', 10, 1) WITH NOWAIT
	UPDATE ProjectVersion SET RelevancyStatus = dbo.RelevancyStatus_NonCleanupable()
	WHERE RelevancyStatus IS NULL AND DeleteRequested = 0 AND (
		-- PV is pinned
		IsPinned = 1
		-- PV's retention time not yet started
		OR RetentionTimeStartDate IS NULL
		-- PV's retention time not yet expired
		OR DATEDIFF(day, RetentionTimeStartDate, CURRENT_TIMESTAMP) <= RetentionTimeInDays
	)
	
	-- Step 5: Determine all PVs that are referenced by non-cleanupable PVs, set RS
	--         on those PVs to NonCleanupable as well
	RAISERROR ('Step 5: Determine all PVs that are referenced by non-cleanupable PVs, set RS  on those PVs to NonCleanupable as well', 10, 1) WITH NOWAIT
	SET @relevancyStatus = dbo.RelevancyStatus_NonCleanupable()
	EXEC dbo.PropagateRelevancyStatus @relevancyStatus
	
	-- Step 6: All other PVs are cleanupable
	RAISERROR ('Step 6: All other PVs are cleanupable', 10, 1) WITH NOWAIT
	UPDATE ProjectVersion SET RelevancyStatus = dbo.RelevancyStatus_Cleanupable() WHERE RelevancyStatus IS NULL
	
	-- Step 7: Rebuild the Dependency.QueryRelevancy level
	RAISERROR ('Step 7: Rebuild the Dependency.QueryRelevancy level', 10, 1) WITH NOWAIT
	EXEC dbo.RebuildQueryRelevancyLevel
END
GO
