PRINT 'CREATING DEPENDDB STATISTICS FUNCTIONS'

-- Drop functions
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'GetRowCounts') AND type in (N'FN', N'IF', N'TF', N'FS', N'FT')) DROP FUNCTION GetRowCounts;
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'GetSizeStatistics') AND type in (N'FN', N'IF', N'TF', N'FS', N'FT')) DROP FUNCTION GetSizeStatistics;
GO

-- GetSizeStatistics
-- Function: Returns a table containing size information about the DependDB tables (the size information is in KB)
CREATE FUNCTION GetSizeStatistics ()
RETURNS 
@SizeStatistics TABLE 
(
	Name nvarchar(128),
	Rows BIGINT,
	Reserved BIGINT,
	Data BIGINT,
	IndexSize BIGINT,
	Unused BIGINT
)
AS
BEGIN
	DECLARE @tableNames CURSOR
	DECLARE @tableName SYSNAME
	DECLARE @exec NVARCHAR
	
	SET @tableNames =
		CURSOR FOR
		SELECT name
		FROM sys.tables
		
	OPEN @tableNames
	FETCH NEXT FROM @tableNames INTO @tableName
	
	WHILE @@FETCH_STATUS = 0 BEGIN
		IF @tableName LIKE 'AspNet_%' BEGIN
			FETCH NEXT FROM @tableNames INTO @tableName
			CONTINUE
		END
	
		DECLARE 
			@objectId INT,
			@pages BIGINT,
			@reservedpages BIGINT,
			@usedpages BIGINT,
			@rowCount BIGINT;
		
		SET @objectId = OBJECT_ID(@tableName,'U')
		
		SELECT   
			@reservedpages = SUM (reserved_page_count),  
			@usedpages = SUM (used_page_count),  
			@pages = SUM (  
				CASE  
					WHEN (index_id < 2) THEN (in_row_data_page_count + lob_used_page_count + row_overflow_used_page_count)  
					ELSE lob_used_page_count + row_overflow_used_page_count  
				END  
			),  
			@rowCount = SUM (  
				CASE  
					WHEN (index_id < 2) THEN row_count  
					ELSE 0  
				END  
			)  
		FROM sys.dm_db_partition_stats  
		WHERE object_id = @objectId;
		
		INSERT INTO @SizeStatistics
		SELECT  
			@tableName as Name,
			@rowCount as Rows,
			@reservedpages * 8 as Reserved,  
			@pages * 8 as Data,  
			(CASE WHEN @usedpages > @pages 
				THEN (@usedpages - @pages) 
				ELSE 0 
			END) * 8 as IndexSize,  
			(CASE WHEN @reservedpages > @usedpages 
				THEN (@reservedpages - @usedpages) 
				ELSE 0 
			END) * 8 as unused
		
		FETCH NEXT FROM @tableNames INTO @tableName
	END
	
	CLOSE @tableNames
	DEALLOCATE @tableNames
	
	RETURN 
END
GO

-- GetRowCounts
-- Function: Returns the number of rows in various DependDB tables matching given project, branch and/or version filters.
--           If the parameters are omitted every project/branch/version is used.
CREATE FUNCTION GetRowCounts 
(
	@projectId int = NULL, 
	@branchId  int = NULL,
	@versionId int = NULL
)
RETURNS 
@Results TABLE 
(
	Projects BIGINT,
	Branches BIGINT,
	Versions BIGINT,
	Assemblies BIGINT,
	Parts BIGINT
)
AS
BEGIN
	DECLARE
		@projects BIGINT,
		@branches BIGINT,
		@versions BIGINT,
		@assemblies BIGINT,
		@parts BIGINT
		
	IF @projectId IS NULL AND @branchId IS NULL AND @versionId IS NULL BEGIN
		SELECT @projects = COUNT(*) FROM Project
		SELECT @branches = COUNT(*) FROM ProjectBranch
		SELECT @versions = COUNT(*) FROM ProjectVersion
		SELECT @assemblies = COUNT(*) FROM Assembly
		SELECT @parts = COUNT(*) FROM DependencyPart
	END ELSE BEGIN
		SET @projects = (
			SELECT COUNT(*)
			FROM Project
			WHERE 
				@projectId IS NULL 
				OR ID = @projectId 
		)

		SET @branches = (
			SELECT COUNT(*)
			FROM ProjectBranch
			WHERE
				(@projectId IS NULL OR ProjectID = @projectId) AND
				(@branchId IS NULL OR ID = @branchId)
		)

		SET @versions = (
			SELECT COUNT(*)
			FROM ProjectVersion v
				INNER JOIN ProjectBranch b ON v.ProjectBranchID = b.ID
			WHERE 
				(@projectId IS NULL OR b.ProjectID = @projectId) AND
				(@branchId IS NULL OR b.ID = @branchId) AND
				(@versionId IS NULL OR v.ID = @versionId)
		)

		SET @assemblies = (
			SELECT COUNT(*)
			FROM ProjectVersionAssembly pva
				INNER JOIN ProjectVersion v ON pva.ProjectVersionID = v.ID
				INNER JOIN ProjectBranch b ON v.ProjectBranchID = b.ID
			WHERE 
				(@projectId IS NULL OR b.ProjectID = @projectId) AND
				(@branchId IS NULL OR b.ID = @branchId) AND
				(@versionId IS NULL OR v.ID = @versionId)
		)

		SET @parts = (
			SELECT COUNT(*)
			FROM DependencyPart p
				INNER JOIN ProjectVersionAssembly pva ON p.AssemblyID = pva.AssemblyID
				INNER JOIN ProjectVersion v ON pva.ProjectVersionID = v.ID
				INNER JOIN ProjectBranch b ON v.ProjectBranchID = b.ID
			WHERE 
				(@projectId IS NULL OR b.ProjectID = @projectId) AND
				(@branchId IS NULL OR b.ID = @branchId) AND
				(@versionId IS NULL OR v.ID = @versionId)
		)
	END

	INSERT INTO @Results
		SELECT 
			@projects as Projects,
			@branches as Branches,
			@versions as Versions,
			@assemblies as Assemblies,
			@parts as Parts

	RETURN 
END
GO
